#!/usr/bin/env python3
"""
This file provides image matching.
"""

# pylint: disable=broad-exception-caught
# pylint: disable=import-outside-toplevel
# pylint: disable=import-error
# pylint: disable=line-too-long
# pylint: disable=no-name-in-module
# pylint: disable=protected-access
# ruff: noqa: E402
# ruff: noqa: F403
# ruff: noqa: E501

from os import path
from time import sleep

from behave import step
from dogtail.rawinput import click

from qecore.utility import RunAndLog, get_func_params_and_values
from qecore.logger import Logging

logging_class = Logging()
LOGGING = logging_class.logger


try:
    import cv2
except ModuleNotFoundError:
    print(
        "You need to install an 'opencv-python' via pip or 'python3-opencv' via yum/dnf."
    )
except ImportError:
    print(
        "A possibility of an error on secondary architecture. Image matching will not be available."
    )

HELP_MESSAGE = """
You are encouraged to build your own step function according to your needs.
Two steps that you see bellow are:
    * General step that just compares and asserts the result.
    * General step that just compares and clicks on the found result.

What is needed for image match:
    * You need to capture an image in which we look for the element you want to find.
        * Provided by capture_image method in Matcher class.
        * This option is True by default.
        * If you have your own, set capture=False and provide self.screen_path in the Matcher class.

    * You need to match the two images, you are looking for a 'needle'.
      So you provide it in function or in step call (.feature file).
        * Provided by match which will return True or False. Lets user react on False return value.
        * Provided by assert_match which will assert the result and terminate the test on False.

    * (Optional) You can draw the result for attachment or
            your own confirmation that matching works.
        * Provided by draw method on Matcher instance to get an image with highlighted needle.
          Highlight is a red rectangle exactly in a place of a match, surrounding provided needle.

    * (Optional) You can click on your found result.
        * Provided by click method in Matcher instance.
        * Requirements are of course success of a match/assert_match.

    * (Optional) You can embed result to test report.
        * For this option the method draw is required.
        * Use method provided in TestSandbox class:
            attach_image_to_report(context, image=image_location, caption="DefaultCaption")
        * Or embed it on your own:
            context.embed("image/png", open(image_location, "rb").read(), caption="DefaultCaption")
        * Remember that result is saved in Matcher instance as
            self.diff_path which equals "/tmp/diff.png"

    * (Optional) You can search only in region of screen
        * Firstly, you have to save the region into context.opencv_region attribute
        * Format of this attribute is tuple (node.position, node.size):
            position and size are (x, y) tuples of integers
"""


@step('Image "{needle}" is shown on screen')
@step('Image "{needle}" is shown on screen with threshold "{threshold:f}"')
@step('Image "{needle}" possibly scaled in range "{scale_range}" is shown on screen with threshold "{threshold:f}"')
def image_match(context, needle, threshold=0.8, scale_range="1,1,1") -> None:
    """
    Function with behave step decorators::

        Image "{needle}" is shown on screen
        Image "{needle}" is shown on screen with threshold "{threshold:f}"
        Image "{needle}" possibly scaled in range "{scale_range}" is shown on screen with threshold "{threshold:f}"

    Explanation:

    * **needle** - location of the file to match.
    * **threshold** - float value of threshold for determination of success/failure.
    * **scale_range** - comma separated triple of numbers: start scale, end scale number of steps, example: "0.8,1.2,21".
    """

    scale = [float(i) for i in scale_range.split(",")]
    scale[-1] = int(scale[-1])
    image_match_instance = Matcher(context)
    image_match_instance.assert_match(needle, threshold, scale=scale)
    image_match_instance.draw()


@step('Image "{needle}" is not shown on screen')
@step('Image "{needle}" is not shown on screen with threshold "{threshold:f}"')
def image_not_match(context, needle, threshold=0.8) -> None:
    """
    Function with behave step decorators::

        Image "{needle}" is not shown on screen
        Image "{needle}" is not shown on screen with threshold "{threshold:f}"

    Explanation:

    * **needle** - location of the file to match.
    * **threshold** - float value of threshold for determination of success/failure.
    """

    image_match_instance = Matcher(context)
    positive_match = image_match_instance.match(needle, threshold)
    if positive_match:
        image_match_instance.draw()
        context.attach_opencv = True
    assert not positive_match, "".join(
        (f"Image '{needle}' was found, that was not supposed to happen.")
    )


@step('Image "{needle}" is shown in region')
@step('Image "{needle}" is shown in region with threshold "{threshold:f}"')
def image_region_match(context, needle, threshold=0.8) -> None:
    """
    Function with behave step decorators::

        Image "{needle}" is shown in region
        Image "{needle}" is shown in region with threshold "{threshold:f}"

    Explanation:

    * **needle** - location of the file to match.
    * **threshold** - float value of threshold for determination of success/failure.
    """

    region = getattr(context, "opencv_region", None)
    assert region is not None, "".join(
        ("No region, you must set context.opencv_region first!")
    )
    image_match_instance = Matcher(context)
    image_match_instance.capture_image()
    image_match_instance.crop_image(region)
    image_match_instance.assert_match(needle, threshold, capture=False)
    image_match_instance.draw()


@step('Image "{needle}" is not shown in region')
@step('Image "{needle}" is not shown in region with threshold "{threshold:f}"')
def image_region_not_match(context, needle, threshold=0.8) -> None:
    """
    Function with behave step decorators::

        Image "{needle}" is not shown in region
        Image "{needle}" is not shown in region with threshold "{threshold:f}"

    Explanation:

    * **needle** - location of the file to match.
    * **threshold** - float value of threshold for determination of success/failure.
    """

    region = getattr(context, "opencv_region", None)
    assert region is not None, "No region, you must set context.opencv_region first!"
    image_match_instance = Matcher(context)
    image_match_instance.capture_image()
    image_match_instance.crop_image(region)
    positive_match = image_match_instance.match(needle, threshold, capture=False)
    if positive_match:
        image_match_instance.draw()
        context.attach_opencv = True
    assert not positive_match, "".join(
        (f"Image '{needle}' was found, that was not supposed to happen.")
    )


@step('Locate and click "{needle}"')
def locate_and_click(context, needle) -> None:
    """
    When used with Wayland, the ponytail will translate the click to the focused window.
    This in effect means the image match found in screenshot will have "global" coordinates.
    Which will be translated by ponytail to the application as a (0,0) instead of the
    screen and when application is not full screen it will be wrong.
    To fix this lets attempt to disconnect the focused window and connect only the monitor.

    With x11 we simply use raw click from dogtail.

    Function with behave step decorator::

        Locate and click "{needle}"

    Explanation:

    * **needle** - location of the file to match.
    """

    LOGGING.debug(get_func_params_and_values())

    image_match_instance = Matcher(context)
    image_match_instance.assert_match(needle)

    # With Wayland lets use ponytail to disconnect focused window as coordinates origin.
    try:
        # Import ponytail to check if it is initiated.
        from dogtail.rawinput import ponytail

        # Checking if ponytail is initiated is a good indicator for Wayland.
        if ponytail is not None:
            LOGGING.debug("Ponytail initiated, coordinates translation not in effect.")

            # Lets disconnect focused window as coordinates origin.
            ponytail.disconnect()
            sleep(1)

            # Lets use the whole screen as coordinates origin.
            ponytail.connectMonitor()

            # Calculate the center for the click.
            match_center_x = (
                image_match_instance.matched_loc[0]
                + int(image_match_instance.needle_width / 2)
                + 5
            )
            match_center_y = (
                image_match_instance.matched_loc[1]
                + int(image_match_instance.needle_height / 2)
                + 5
            )

            # Click with rawinput to exact coordinates (some small offset required).
            LOGGING.debug(
                f"Coordinates for ponytail: ({match_center_x}, {match_center_y})"
            )
            ponytail.generateButtonEvent(1, match_center_x, match_center_y)
            return

    except Exception as error:
        LOGGING.info(f"Unexpected error: '{error}'")

    # With Xorg we do not need to change anything and lets do simple rawinput click.
    LOGGING.debug("Clicking to the needle coordinates.")
    image_match_instance.click()


class Matcher:
    """
    Matcher class.
    """

    def __init__(self, context) -> None:
        """
        Initiate Matcher instance.

        :type context: <behave.runner.Context>
        :param context: Context object.
        """

        self.context = context
        self.screen_path = "/tmp/pic.png"
        self.diff_path = "/tmp/diff.png"
        if RunAndLog("which gnome-screenshot").return_code == 0:
            self.capture_image_cmd = "gnome-screenshot -f "
        else:
            self.capture_image_cmd = "qecore_capture_screenshot "
        self.needle_width = 0
        self.needle_height = 0
        self.matched_value = 0.0
        self.matched_loc = (0, 0)

        self.ori_img = None
        self.ori_img_gray = None
        self.needle = None
        self.needle_size = None

    def capture_image(self, screen_path=None) -> None:
        """
        Captures the image.

        :type screen_path: Path to store the screenshot.
        :param screen_path: str.
        """

        LOGGING.debug(get_func_params_and_values())

        if screen_path is None:
            screen_path = self.screen_path
        remove_screen_run = RunAndLog(f"rm -rfv {screen_path}")
        LOGGING.debug(
            f"Image matching remove older image: '{remove_screen_run.output}'"
        )

        capture_image_run = RunAndLog(self.capture_image_cmd + screen_path, timeout=5)
        if capture_image_run.output.startswith("file:"):
            screen_captured = capture_image_run.output.replace("file://", "").strip(
                "\n"
            )
            copy_screen_run = RunAndLog(f"cp {screen_captured} {screen_path}")
            LOGGING.debug(
                f"Image matching copy captured image: '{copy_screen_run.output}'"
            )

        LOGGING.debug(
            f"Image matching initial screen capture result: '{capture_image_run.output}'"
        )

    def crop_image(self, region) -> None:
        """
        Crops the image.

        :type region: list
        :param region: List with four values: (x, y, w, h)
        """

        LOGGING.debug(get_func_params_and_values())

        self.ori_img = cv2.imread(self.screen_path)
        (x, y), (w, h) = region
        if self.context.sandbox.session_type == "wayland":
            y += self.context.sandbox.shell.child("System", "menu").size[1]
        self.ori_img = self.ori_img[y : y + h, x : x + w]
        cv2.imwrite(self.screen_path, self.ori_img)

    def assert_match(
        self, needle, threshold=0.8, capture=True, scale=(1.0, 1.0, 1)
    ) -> None:
        """
        Calls and asserts the result of :py:func:`match`.

        :type needle: str
        :param needle: Needle location.

        :type threshold: float
        :param threshold: Value of acceptable match.

        :type scale: tuple
        :param scale: Triple representing scale range: (scale start, scale end, step count)

        :type capture: bool
        :param capture: Decides if the image will be captured.
        """

        LOGGING.debug(get_func_params_and_values())

        assert self.match(needle, threshold, capture, scale), "".join(
            (f"Image match value: {self.matched_value}")
        )

    def match(self, needle, threshold=0.8, capture=True, scale=(1.0, 1.0, 1)) -> bool:
        """
        Trying to find the needle image inside the captured image.

        :type needle: str
        :param needle: Needle location.

        :type threshold: float
        :param threshold: Value of acceptable match.

        :type capture: bool
        :param capture: Decides if the image will be captured.

        :type scale: tuple
        :param scale: Triple representing scale range: (scale start, scale end, step count)

        :rtype: bool
        :return: Boolean value of the matching.
        """

        LOGGING.debug(get_func_params_and_values())

        if capture:
            self.capture_image()

        self.ori_img = cv2.imread(self.screen_path)
        self.ori_img_gray = cv2.cvtColor(self.ori_img, cv2.COLOR_BGR2GRAY)
        self.needle = cv2.imread(path.abspath(needle), 0)
        best_match = 0
        scale_start, scale_end, scale_count = scale
        # avoid division by zero
        scale_step = (
            0 if scale_count == 1 else (scale_end - scale_start) / (scale_count - 1)
        )
        for i in range(scale_count):
            scale_factor = scale_start + i * scale_step
            # for some reason, cv2.resize destination size is reversed shape
            new_size = tuple(
                reversed([int(x * scale_factor) for x in self.needle.shape[:2]])
            )
            resized = cv2.resize(self.needle, new_size, interpolation=cv2.INTER_AREA)
            match = cv2.matchTemplate(self.ori_img_gray, resized, cv2.TM_CCOEFF_NORMED)
            _, matched_value, _, matched_loc = cv2.minMaxLoc(match)
            if matched_value > best_match:
                best_match = matched_value
                self.matched_value = matched_value
                self.matched_loc = matched_loc
                self.needle_width, self.needle_height = resized.shape[::-1]

        if hasattr(self.context, "sandbox") and (
            self.context.sandbox.attach_screenshot_on_pass
            or self.context.sandbox._embed_all
        ):
            self.context.sandbox.attach_image_to_report(self.context, needle, "Needle")
            self.context.sandbox.attach_image_to_report(
                self.context, self.screen_path, "Searching in Screenshot"
            )

        return self.matched_value > threshold

    def draw(self) -> None:
        """
        Draws the result to the original image.
        """

        LOGGING.debug(get_func_params_and_values())

        self.needle_size = (
            self.matched_loc[0] + self.needle_width,
            self.matched_loc[1] + self.needle_height,
        )
        cv2.rectangle(self.ori_img, self.matched_loc, self.needle_size, (0, 0, 255), 2)
        cv2.imwrite(self.diff_path, self.ori_img)
        if hasattr(self.context, "sandbox") and (
            self.context.sandbox.attach_screenshot_on_pass
            or self.context.sandbox._embed_all
        ):
            self.context.sandbox.attach_image_to_report(
                self.context, self.diff_path, "Matched Screenshot"
            )

    def click(self) -> None:
        """
        Clicks to the center of the result.
        """

        LOGGING.debug(get_func_params_and_values())

        match_center_x = self.matched_loc[0] + int(self.needle_width / 2)
        match_center_y = self.matched_loc[1] + int(self.needle_height / 2)

        LOGGING.debug(f"Clicking at coordinates: ({match_center_x},{match_center_y})")
        click(match_center_x, match_center_y)
        sleep(1)


def debug_click_via_image(context, atspi_object, click_button=1):
    """
    Debug function to take a screenshot and using opencv insert click position.

    Keep in mind that ponytail will translate coordinates to the focused window on Wayland.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :param atspi_object: Atspi object to be clicked.
    :type atspi_object: <Accessible...>

    :param click_button: Mouse button to click, defaults to 1
    :type click_button: int
    """

    LOGGING.debug(get_func_params_and_values())
    LOGGING.debug("Ponytail will translate coordinates to the focused window.")

    # Make a screenshot and load base file as base image.
    base_image = "/tmp/test.png"
    # Make a new screenshot.
    Matcher(context).capture_image(base_image)
    sleep(0.5)

    # Load the screenshot as a base.
    grid_image_target = cv2.imread(base_image)
    click_point_image_target = cv2.imread(base_image)

    starting_point = 0

    increment_small = 20
    starting_point_small = 20

    ending_point = 10000

    # Make first debug image, grid of lines over the base image.
    grid_debug_image = "/tmp/grid_debug_image.png"
    LOGGING.debug(f"File for grid debug: '{grid_debug_image}'")

    for i in range(100):
        v_start_point = (starting_point, starting_point_small + increment_small * i)
        v_end_point = (ending_point, starting_point_small + increment_small * i)
        start_point = (starting_point_small + increment_small * i, starting_point)
        end_point = (starting_point_small + increment_small * i, ending_point)

        if (starting_point_small + increment_small * i) % 100 == 0:
            color = (106, 212, 212)
            thickness = 2
        else:
            color = (78, 109, 153)
            thickness = 1

        grid_image_target = cv2.line(
            grid_image_target, start_point, end_point, color, thickness
        )
        grid_image_target = cv2.line(
            grid_image_target, v_start_point, v_end_point, color, thickness
        )

    if not hasattr(context, "grid_shown") or (
        hasattr(context, "grid_shown") and not context.grid_shown
    ):
        LOGGING.debug("Writing grid_debug_image")
        cv2.imwrite(grid_debug_image, grid_image_target)
        context.grid_shown = True

    # Second debug image based on click position.
    click_point_debug_image = "/tmp/click_point_debug_image.png"
    LOGGING.debug(f"File for click debug: '{click_point_debug_image}'")

    x = int(atspi_object.position[0] + atspi_object.size[0] / 2)
    y = int(atspi_object.position[1] + atspi_object.size[1] / 2)
    cv2.circle(
        click_point_image_target, (x, y), radius=0, color=(0, 0, 255), thickness=8
    )
    cv2.circle(
        click_point_image_target, (x, y), radius=0, color=(255, 255, 255), thickness=5
    )
    cv2.putText(
        click_point_image_target,
        f"({x},{y})",
        (x, y),
        cv2.FONT_HERSHEY_SIMPLEX,
        1,
        (211, 0, 160),
        1,
    )

    LOGGING.debug("Writing click_point_debug_image")
    cv2.imwrite(click_point_debug_image, click_point_image_target)

    LOGGING.debug(f"Clicking at x:'{x}', y:'{y}' to obj {atspi_object.name}")

    try:
        atspi_object.click(click_button)
    except Exception:
        click(x, y, click_button)

    context.embed(
        mime_type="image/png",
        data=grid_debug_image,
        caption="Grid Image",
    )
    cap = f"Click ({x}, {y})"
    context.embed(
        mime_type="image/png",
        data=click_point_debug_image,
        caption=cap,
    )
    sleep(1)



#!/usr/bin/env python3
"""
QECore __init__ file.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = """Michal Odehnal <modehnal@redhat.com>"""
__copyright__ = "Copyright © 2018-2025 Red Hat, Inc."
__license__ = "GPLv3"
__all__ = (
    "api",
    "application",
    "common_steps",
    "flatpak",
    "get_node",
    "image_matching",
    "logger",
    "online_accounts",
    "sandbox",
    "utility",
    "icons",
)



#!/usr/bin/env python3
"""
Common steps part with definition of pre-coded decorator steps for behave usage.
"""

# pylint: disable=broad-exception-caught
# pylint: disable=import-outside-toplevel
# pylint: disable=import-error
# pylint: disable=invalid-name
# pylint: disable=no-name-in-module
# pylint: disable=line-too-long
# ruff: noqa: E501

from time import sleep
import subprocess

from behave import step

from dogtail.rawinput import (
    click,
    keyCombo,
    pressKey,
    typeText,
    absoluteMotion,
)

from qecore.utility import get_application
from qecore.get_node import GetNode
from qecore.step_matcher import use_step_matcher


__author__ = """
Filip Pokryvka <fpokryvk@redhat.com>,
Michal Odehnal <modehnal@redhat.com>,
Bohdan Milar <bmilar@redhat.com>
"""

# To stop using "qecore" matcher use matcher "parse".
use_step_matcher("qecore")

HELP = """
This allows multiple decorator definitions on one line separated by ' | '
EXAMPLE:
This decorator:
    @step('Item "{name}" "{role_name}" | with description "{description}" | that is "{attr}"')

matches for example the following steps:
    * Item "foo" "push button"
    * Item "foo" "push button" with description "something useful"
    * Item "foo" "push button" with description "something useful" that is "visible"
    * Item "foo" "push button" that is "visible"

And also any permutation of decorator parts except the first part, so this is also valid step:
    * Item "foo" "push button" that is "visible" with description "something useful"

WARNING:
"qecore" matcher does not work well with unquoted decorator arguments, use "parse" instead
"""

SIZE_DEC = "".join([' | has size at least "{size_low}"',
                    ' | has size at most "{size_high}"'])
POS_DEC = "".join([' | that is at least "{position_low}" from top left',
                   ' | that is at most "{position_high}" from top left'])
SIZE_POS_DEC = SIZE_DEC + POS_DEC

DESC = ' | with description "{description}"'
ACTION = ' | with action "{action}"'
ATTR = ' | that is "{attr}" | is "{attr}"'
ROOT = ' | in "{a11y_root_name}"'
TEXT = ' | has text "{text}"'

NOT_ATTR = ' | that is not "{attr}" | is not "{attr}"'
NOT_TEXT = ' | does not have text "{text}"'
NOT_DESC = ' | does not have description "{description}"'

START_OPTIONS = ' | via "{start_via}" | via command "{command}" | via command | in "{session}" | with env "{environ}"'

# Format ' | with offset "(0,0)"'
OFFSET = ' | with offset "{offset}"'


@step('{m_btn} click "{name}" "{role_name}"' + DESC + ACTION + OFFSET + ATTR + ROOT + SIZE_POS_DEC)
def mouse_click(context, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Mouse clicking function to be used by behave feature files.

    This step will always use raw click.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (data, node):
        try:
            # Newer dogtail has window_id.
            click(
                x=data.x_node_center,
                y=data.y_node_center,
                button=data.m_btn,
                window_id=node.window_id
            )
        except (AttributeError, TypeError):
            # Older dogtail does not have window_id.
            click(x=data.x_node_center, y=data.y_node_center, button=data.m_btn)


@step('Mouse over "{name}" "{role_name}"' + DESC + ACTION + OFFSET + ATTR + ROOT + SIZE_POS_DEC)
def mouse_over(context, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Mouse over function to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (data, node):
        try:
            # Newer dogtail has window_id.
            absoluteMotion(
                x=data.x_node_center,
                y=data.y_node_center,
                window_id=node.window_id
            )
        except (AttributeError, TypeError):
            # Older dogtail does not have window_id.
            absoluteMotion(x=data.x_node_center,y=data.y_node_center)


@step('Make an action "{action}" for "{name}" "{role_name}"' + DESC + ACTION + ATTR + ROOT + SIZE_POS_DEC)
def make_action(context, action, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Accessibility action function to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        node.doActionNamed(action)


@step('Item "{name}" "{role_name}" | found |' + DESC + ACTION + ATTR + ROOT + SIZE_POS_DEC)
def node_attribute(context, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Verification function to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        assert node is not None, "Node was not found, it should be!"


@step('Item "{name}" "{role_name}" | was not found |' + DESC + ACTION + NOT_ATTR + ROOT + SIZE_POS_DEC)
def node_not_attribute(context, retry=False, expect_positive=False, **kwargs) -> None:
    """
    Verification function negated to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        assert node is None, "Node was found, it should not be!"


@step('Item "{name}" "{role_name}"' + DESC + ACTION + TEXT + ROOT + SIZE_POS_DEC)
def node_with_text(context, text, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Verification function for text attribute to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        assert text in node.text, "".join((
            f"Found node should have text: {text}\n",
            f"Instead the node has text: {node.text}",
        ))


@step('Item "{name}" "{role_name}"' + DESC + ACTION + NOT_TEXT + ROOT + SIZE_POS_DEC)
def node_without_text(context, text, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Verification function for text attribute negated to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        assert text not in node.text, "".join((
            f"Found node should have text: {text}\n",
            f"Node was found with text: {node.text}"
        ))


@step('Item "{name}" "{role_name}"' + NOT_DESC + ACTION + ATTR + ROOT + SIZE_POS_DEC)
def node_without_description(context, description, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Verification function for description attribute negated to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        assert description not in node.description, "".join((
            f"Found node should not have description: {description}\n",
            f"Instead the node has description: {node.description}"
        ))


@step('Wait until "{name}" "{role_name}"' + DESC + ACTION + ATTR + ROOT + SIZE_POS_DEC)
def wait_until_attr(context, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Wait until accessibility attribute becomes true function to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        for _ in range(30):
            if not getattr(node, kwargs.get("attr", "sensitive")):
                sleep(0.2)
            else:
                return


@step('Wait until "{name}" "{role_name}" | appears' + DESC + ACTION + ROOT + SIZE_POS_DEC)
def wait_until_in_root(context, retry=True, expect_positive=True, **kwargs) -> None:
    """
    Wait until accessibility object appears function to be used by behave feature files.
    """

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        if node is not None:
            return


@step('Wait until "{name}" "{role_name}" | disappears' + DESC + ACTION + NOT_ATTR + ROOT + SIZE_POS_DEC)
def wait_until_not_in_root(context, retry=False, expect_positive=False, **kwargs) -> None:
    """
    Wait until accessibility object disappears function to be used by behave feature files.
    """

    for _ in range(10):
        with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
            if node is None:
                return
        sleep(1)

    with GetNode(context, **kwargs, retry=retry, expect_positive=expect_positive) as (_, node):
        assert node is None, "Node still found in a11y tree."


@step('Start another instance of "{application_name}"' + START_OPTIONS)
def start_another_instance_of_application(
    context,
    application_name=None,
    start_via="command",
    command=None,
    session=None,
    kill=False
) -> None:
    """
    Start another application instance function to be used by behave feature files.
    """

    start_application(context, application_name, start_via, command, session, kill)


@step('Start application "{application_name}"' + START_OPTIONS)
def start_application(
    context,
    application_name=None,
    start_via="command",
    command=None,
    session=None,
    environ=None,
    kill=True
) -> None:
    """
    Start application function to be used by behave feature files.
    """

    application = get_application(context, application_name)
    if start_via == "menu":
        try:
            application.start_via_menu(kill=kill)
        except Exception:  # pylint: disable=broad-except
            application.start_via_menu(kill=kill)
    elif start_via == "command":
        try:
            application.start_via_command(command=command, in_session=session, kill=kill, environ=environ)
        except RuntimeError as error:
            assert False, error
        except Exception:  # pylint: disable=broad-except
            application.start_via_command(command=command, in_session=session, kill=kill, environ=environ)
    else:
        raise AssertionError("Only defined options are 'command' and 'menu'.")


# Stop using qecore matcher.
use_step_matcher("parse")


@step('Commentary')
def commentary_step(context) -> None:
    """
    Commentary step for usage in behave feature files - html-pretty only.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>
    """

    # Defined only for html-pretty formatter.
    # This will return an instance of PrettyHTMLFormatter.
    formatter_instance = getattr(context, "html_formatter", None)
    if formatter_instance is not None and formatter_instance.name == "html-pretty":
        # Get the correct step to override.
        scenario = formatter_instance.current_scenario
        # Current scenario is never none, as this step is being executed.
        scenario_step = scenario.current_step

        # Override the step, this will prevent the decorator to be generated and only
        # the text will show.
        scenario_step.set_commentary(True)


@step('Close application "{application_name}" via "{close_via}"')
def application_in_not_running(context, application_name=None, close_via="gnome panel") -> None:
    """
    Close application function to be used by behave feature files.
    """

    application = get_application(context, application_name)

    if close_via == "gnome panel":
        # Menu is not in GNOME panel from rhel-10.
        if "10." in context.sandbox.distribution_version:
            assert False, "Application is not present in gnome panel on rhel-10."

        gnome_panel = context.sandbox.shell.child(application.name, "menu")
        for _ in range(10):
            gnome_panel.click()
            sleep(0.5)
            gnome_menu = context.sandbox.shell.findChildren(
                lambda x: x.name == "Quit" and x.roleName == "label"
            )
            if gnome_menu != []:
                gnome_menu[0].click()
                return
            pressKey("Esc")

        assert False, f"Unable to close '{application_name}' via '{close_via}' in 10 tries."

    elif close_via == "gnome panel with workaround":
        # Menu is not in GNOME panel from rhel-10.
        if "10." in context.sandbox.distribution_version:
            assert False, "Application is not present in gnome panel on rhel-10."

        gnome_panel = context.sandbox.shell.child(application.name, "menu")
        for _ in range(10):
            gnome_panel.click()
            sleep(0.5)
            gnome_menu = context.sandbox.shell.findChildren(
                lambda x: x.name == "Quit" and x.roleName == "label"
            )
            if gnome_menu != []:
                gnome_menu[0].point()
                sleep(0.5)
                pressKey("Enter")
                return
            pressKey("Esc")

        assert False, f"Unable to close '{application_name}' via '{close_via}' in 10 tries."

    elif (close_via == "application panel menu") or (
        close_via == "application menu" and context.sandbox.distribution == "Fedora"
    ):
        application_panel = application.instance.children[0][0]
        for _ in range(10):
            application_panel.click(3)
            sleep(0.5)
            application_panel_menu = context.sandbox.shell.findChildren(
                lambda x: x.name == "Close" and x.roleName == "label"
            )
            if application_panel_menu != []:
                application_panel_menu[0].click()
                return
            pressKey("Esc")

        assert False, f"Unable to close '{application_name}' via '{close_via}' in 10 tries."

    elif close_via in ("application file menu", "application menu"):
        for _ in range(10):
            context.execute_steps(f'* Left click "File" "menu" in "{application.component}"')
            sleep(0.5)
            application_file_menu = application.instance.findChildren(
                lambda x:
                ("Close" in x.name or "Quit" in x.name)
                and x.roleName == "menu item" and x.sensitive
            )
            if application_file_menu != []:
                application_file_menu[0].click()
                return
            pressKey("Esc")

        assert False, f"Unable to close '{application_name}' via '{close_via}' in 10 tries."

    elif close_via == "application toggle menu":
        for _ in range(10):
            context.execute_steps(f'* Left click "Application menu" "toggle button" in "{application.component}"')
            sleep(0.5)
            try:
                context.execute_steps(f'* Left click "Quit" "push button" in "{application.component}"')
                return
            except Exception:  # pylint: disable=broad-except
                pressKey("Esc")

        assert False, f"Unable to close '{application_name}' via '{close_via}' in 10 tries."

    elif "menu:" in close_via:
        menu_name = None
        menu_item_name = None
        try:
            _, menu_name, menu_item_name = close_via.split(":")
        except Exception as error:
            raise UserWarning(
                f"Expected format of specific close via menu is 'menu:<menu>:<menu_item>\n{error}"
            ) from error

        for _ in range(10):
            application.instance.child(menu_name, "menu").click()
            sleep(0.5)
            try:
                application.instance.child(menu_item_name, "menu item").click()
                return
            except Exception:  # pylint: disable=broad-except
                pressKey("Esc")

        assert False, f"Unable to close '{application_name}' via '{close_via}' in 10 tries."

    elif close_via == "shortcut":
        application.close_via_shortcut()

    elif close_via == "kill command":
        application.kill_application()

    else:
        raise AssertionError("".join((
            "Only defined options are:\n",
            "'gnome panel', 'application menu', 'shortcut', 'kill command', \n",
            "'application file menu', 'application toggle menu' and 'application panel menu'."
        )))


@step('Application "{application_name}" is no longer running')
def application_is_not_running(context, application_name) -> None:
    """
    Application is not running function to be used by behave feature files.
    """

    application = get_application(context, application_name)
    if application.is_running():
        application.wait_before_app_closes(15)


@step('Application "{application_name}" is running')
def application_is_running(context, application_name) -> None:
    """
    Application is running function to be used by behave feature files.
    """

    application = get_application(context, application_name)
    application.already_running()
    if not application.is_running():
        application.wait_before_app_starts(15)


@step('Run and save command output: "{command}"')
def run_and_save(context, command) -> None:
    """
    Run a shell command and store its returncode, stdout and stderr to context.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :type command: str
    :param command: Command line to be executed and result stored.
    """

    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        encoding="utf-8",
    )
    try:
        stdout, stderr = process.communicate(timeout=5)
    except subprocess.TimeoutExpired:
        process.terminate()
        stdout, stderr = process.communicate()

    context.command = command
    context.command_stdout = stdout
    context.command_stderr = stderr
    context.command_return_code = process.returncode


@step('Last command output "{operand}" "{expected_output}"')
def verify_content_in_output(context, operand, expected_output) -> None:
    """
    Verify that the result of last command has a certain data as a result.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :type operand: str
    :param operand: String specifying what operation to do.

    :type expected_output: str
    :param expected_output: Expected data to be compared to the command output.
    """

    assert hasattr(context, "command_stdout"), "".join((
        "\nYou have not saved a command output to be checked.",
        "\nTo do that use '* Run and save command output: \"<command>\"'"
    ))

    valid_operands = ("is", "is not", "contains", "does not contain", "begins with",
                      "does not begin with", "ends with", "does not end with")
    assert operand in valid_operands, "".join((
        f"You have attempted to use operand: '{operand}'",
        f"But only defined operands are:\n'{valid_operands}'"
    ))

    command_output = context.command_stdout
    if operand == "is":
        assert expected_output == command_output, "".join((
            f"\nWanted output: '{expected_output}'",
            f"\nActual output: '{command_output}'"
        ))

    if operand == "is not":
        assert expected_output != command_output, "".join((
            f"\nNot Wanted output: '{expected_output}'",
            f"\nActual output: '{command_output}'"
        ))

    if operand == "contains":
        assert expected_output in command_output, "".join((
            f"\nOutput should contain: '{expected_output}'",
            f"\nBut it does not: '{command_output}'"
        ))

    if operand == "does not contain":
        assert expected_output not in command_output, "".join((
            f"\nOutput should not contain: '{expected_output}'",
            f"\nBut it does: '{command_output}'"
        ))

    if operand == "begins with":
        assert command_output.startswith(expected_output), "".join((
            f"\nOutput should begin with: '{expected_output}'",
            f"\nBut it does not: '{command_output}'"
        ))

    if operand == "does not begin with":
        assert not command_output.startswith(expected_output), "".join((
            f"\nOutput should not begin with: '{expected_output}'",
            f"\nBut it does: '{command_output}'"
        ))

    if operand == "ends with":
        assert command_output.endswith(expected_output), "".join((
            f"\nOutput should end with: '{expected_output}'",
            f"\nBut it does not: '{command_output}'"
        ))

    if operand == "does not end with":
        assert not command_output.endswith(expected_output), "".join((
            f"\nOutput should not end with: '{expected_output}'",
            f"\nBut it does: '{command_output}'"
        ))


@step('Return code of last command output "{operand}" "{expected_return_code}"')
def verify_return_code_of_command(context, operand, expected_return_code) -> None:
    """
    Verify the return code of last command.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :type operand: str
    :param operand: String specifying what operation to do.

    :type expected_return_code: str
    :param expected_return_code: Expected data to be compared to the command output.
    """

    assert hasattr(context, "command_stdout"), "".join((
        "\nYou have not saved a command output to be checked.",
        "\nTo do that use '* Run and save command output: \"<command>\"'"
    ))

    valid_operands = ("is", "is not")
    assert operand in valid_operands, "".join((
        f"You have attempted to use operand: '{operand}'",
        f"But only defined operands are:\n'{valid_operands}'"
    ))

    return_code = context.command_return_code
    if operand == "is":
        assert int(expected_return_code) == int(return_code), "".join((
            f"\nWanted return code: '{expected_return_code}'",
            f"\nActual return code: '{return_code}'"
        ))

    if operand == "is not":
        assert int(expected_return_code) != int(return_code), "".join((
            f"\nNot Wanted return code: '{expected_return_code}'",
            f"\nActual return code: '{return_code}'"
        ))


@step('Type text: "{text}"')
def type_text_step(context, text) -> None:  # pylint: disable=unused-argument
    """
    Typing text function to be used by behave feature files.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :param text: Text to type.
    :type text: str
    """

    typeText(text)


@step('Press key: "{key_name}"')
def press_key_step(context, key_name) -> None:  # pylint: disable=unused-argument
    """
    Press key function to be used by behave feature files.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :param key_name: Key to press.
    :type key_name: str
    """

    pressKey(key_name)


@step('Key combo: "{combo_name}"')
def key_combo_step(context, combo_name) -> None:  # pylint: disable=unused-argument
    """
    Key combo function to be used by behave feature files.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :param combo_name: Combo to execute.
    :type combo_name: str
    """

    keyCombo(combo_name)


@step('Wait {number} second before action')
@step('Wait {number} seconds before action')
def wait_up(context, number) -> None:  # pylint: disable=unused-argument
    """
    Sleep function to be used by behave feature files.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :param number: Numeric value for sleep.
    :type number: str
    """

    sleep(int(number))


@step('Move mouse to: x: "{position_x}", y: "{position_y}"')
def absolute_motion_step(context, position_x, position_y) -> None:  # pylint: disable=unused-argument
    """
    Move mouse to coordinates function to be used by behave feature files.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :param position_x: Position X to move mouse
    :type position_x: str

    :param position_y: Position Y to move mouse to.
    :type position_y: str
    """

    absoluteMotion(int(position_x), int(position_y))


@step('{button} click on: x: "{position_x}", y: "{position_y}"')
def click_on_position(context, button, position_x, position_y) -> None:  # pylint: disable=unused-argument
    """
    Mouse click to coordinates function to be used by behave feature files.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>

    :param button: Mouse button to press.
    :type button: str

    :param position_x: Position X to click.
    :type position_x: str

    :param position_y: Position Y to click.
    :type position_y: str
    """

    buttons = {"Left": 1, "Middle": 2, "Right": 3}
    click(int(position_x), int(position_y), buttons[button])


@step('Dummy pass')
def dummy_pass(context):  # pylint: disable=unused-argument
    """
    Dummy step to always PASS.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>
    """

    assert True, "This step always passes."


@step('Dummy fail')
def dummy_fail(context):  # pylint: disable=unused-argument
    """
    Dummy step to always FAIL.

    :param context: Holds contextual information during the running of tests.
    :type context: <behave.runner.Context>
    """

    assert False, "This test always fails."



#!/usr/bin/env python3
"""
Flatpak class to be used from TestSandbox to keep list of running applications.
"""

__author__ = """
Martin Krajnak <mkrajnak@redhat.com>
"""

from os import walk, path
import configparser

from qecore.utility import RunAndLog, get_func_params_and_values

from qecore.application import Application

from qecore.logger import Logging

logging_class = Logging()
LOGGING = logging_class.logger


class Flatpak(Application):
    """
    Flatpak specific class.

    :param Application: Application class.
    :type Application: Application
    """

    def __init__(self, flatpak_id, **kwargs) -> None:
        """
        Initiate Flatpak instance, inherits Application.

        :type flatpak_id: str
        :param flatpak_id: Unique flatpak identifier, mandatory format has 2 dots.
            Param is passed to Application constructor as .component

        :param kwargs: :py:class:`application.Application` parameters.
        """

        LOGGING.debug(get_func_params_and_values())

        # Check if the input is a list of IDs to check.
        if isinstance(flatpak_id, list):
            for flatpak_id_item in flatpak_id:
                try:
                    super().__init__(component=flatpak_id_item, **kwargs)
                    return  # Return successful match.
                except UserWarning:
                    LOGGING.debug(f"Attempt for ID '{flatpak_id_item}' unsuccessful.")

            # If the list was provided, do not continue on string handling.
            RunAndLog("flatpak list")
            raise UserWarning(f"None of the provided IDs '{flatpak_id}' were accepted.")

        # Check the flatpak name if user provided a string.
        if flatpak_id.count(".") != 2:
            raise UserWarning(
                f"Incorrect flatpak name {flatpak_id}, e.g.: 'org.gnome.gedit'."
            )

        # Attempt to check and log if the user did not do a typo.
        try:
            super().__init__(component=flatpak_id, **kwargs)
        except UserWarning as warning:
            LOGGING.debug(f"Flatpak ID provided by the user: '{flatpak_id}' not found.")
            RunAndLog("flatpak list")
            raise UserWarning(f"Flatpak ID '{flatpak_id}' was not found.") from warning

    def kill_application(self) -> None:
        """
        Killing via 'flatpak kill <flatpak_id>', sudo for @system flatpaks.
        """

        LOGGING.debug(get_func_params_and_values())

        if self.is_running() and self.kill:
            # Kills a flatpak installed @System.
            RunAndLog(f"sudo flatpak kill {self.component}")
            # Kills a flatpak installed @User.
            RunAndLog(f"flatpak kill {self.component}")

    def get_desktop_file_data(self) -> None:
        """
        Provide information from .desktop file, two possible locations:
            * flatpak installed as *--user*::

                ~/.local/share/flatpak/app/<flatpak_id>/<arch>/....

            * flatpak installed as *@system* (sudo, root)::

                /var/lib/flatpak/app/<flatpak_id>/<arch>/.....

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`__init__`.
        """

        LOGGING.debug(get_func_params_and_values())

        def get_desktop_file_path(flatpak_dir) -> None:
            """
            Helper function used for dynamic search of desktop files as
            the path contains a commit number in a folder name.
            """

            desktop_file_name = self.desktop_file_name or self.component
            for root, _, files in walk(flatpak_dir):
                for _file in files:
                    if ".desktop" in _file and desktop_file_name in _file:
                        return path.join(root, _file)
            return None

        if not self.desktop_file_exists:
            return None

        desktop_file = None
        if self.desktop_file_path:
            desktop_file = self.desktop_file_path
        else:
            for pth in ["~/.local/share/flatpak/app/", "/var/lib/flatpak/app/"]:
                pth = path.expanduser(pth)
                if path.isdir(f"{pth}{self.component}"):
                    desktop_file = get_desktop_file_path(f"{pth}{self.component}")
                    break

        if not desktop_file:
            raise UserWarning(f"Desktop file for {self.component} not found.")

        desktop_file_config = configparser.RawConfigParser()
        desktop_file_config.read(desktop_file)

        self.name = desktop_file_config.get("Desktop Entry", "name")
        self.exec = f"flatpak run {self.component}"
        self.icon = desktop_file_config.get("Desktop Entry", "icon", fallback=None)

    def is_running(self) -> bool:
        """
        Double check if running application is really a flatpak.
        """

        LOGGING.debug(get_func_params_and_values())

        flatpak_is_running = self.component in RunAndLog("flatpak ps").output

        if not flatpak_is_running:
            LOGGING.debug(f"Flatpak '{self.name}' is not running.")

        return super().is_running() and flatpak_is_running

    def start_via_menu(self, kill=False) -> None:
        """
        Icons belonging to flatpak applications cannot be distinguished
        from icons of their rpm versions, thus we recommend to remove
        rpm versions of the tested applications during test runs.

        .. note::

            Make sure that the .rpm version of a tested flatpak is not installed.
        """

        LOGGING.debug(get_func_params_and_values())

        rpms_run = RunAndLog(f'rpm -qa | grep -i "{self.name}"')
        if rpms_run.output is not None:
            print(
                f"WARNING: a method could possibly start a non-flatpak application\n"
                f"{self.name} is already installed as .rpm: {rpms_run.output}"
            )

        super().start_via_menu(kill)

    def get_pid_list(self) -> None:
        """
        Not required, killing via flatpak kill <flatpak_id>.

        .. note::

            This method is not available for flatpak objects.
        """

        LOGGING.debug(get_func_params_and_values())

        raise NotImplementedError("Not available for flatpak objects.")

    def get_all_kill_candidates(self) -> None:
        """
        Not required, killing via flatpak kill <flatpak_id>.

        .. note::

            This method is not available for flatpak objects.
        """

        LOGGING.debug(get_func_params_and_values())

        raise NotImplementedError("Not available for flatpak objects.")



#!/usr/bin/env python3
"""
Definition of "qecore" matcher.
This matcher splits decorators by '|' to allow all combinations in one line

USAGE::

    from qecore.step_matcher import use_step_matcher
    use_step_matcher("qecore")

    ... # definition of steps

    use_step_matcher("parse") # stop using qecore matcher
"""

# pylint: disable=import-outside-toplevel

from behave.matchers import ParseMatcher
from behave.model_core import Argument

__author__ = """
Filip Pokryvka <fpokryvk@redhat.com>
"""


def use_step_matcher(matcher_name) -> None:
    """
    Overrides `behave.matchers.use_step_matcher()` function.
    Appends "qecore" matcher to behave matchers and selects matcher to be used.

    :type matcher_name: str
    :param matcher_name: name of behave matcher to be used
    """

    from behave.matchers import (
        matcher_mapping,
    )
    from behave.matchers import (
        use_step_matcher as usm,
    )

    matcher_mapping["qecore"] = QecoreMatcher
    usm(matcher_name)


class QecoreMatcher(ParseMatcher):
    """
    Uses :class:`~ParseMatcher` with additional '|' parsing
    """

    delimiter = "|"
    """
    Split delimiter (default "|").

    :type delimiter: str
    """

    start_phrase = "#__start__#"
    """
    Used internally to match beginning of the step.

    :type start_phrase: str
    """

    def __init__(self, func, pattern, step_type=None) -> None:
        """
        Initiate :class:`~QecoreMatcher` instance.
        Split pattern by delimiter (default "|").
        Called by behave.

        :type func: <function>
        :param func: Step function.

        :type pattern: str
        :param pattern: Decorator of step function.

        :type step_type: str
        :param step_type: Type of behave step.
        """

        super(QecoreMatcher, self).__init__(func, pattern, step_type)
        # Note the space at the end of patterns (to make sure whole words are matched).
        self.patterns = [
            f"{self.start_phrase}{p.strip()} " for p in pattern.split(self.delimiter)
        ]
        self.parsers = {}
        for _pattern in self.patterns:
            self.parsers[_pattern] = self.parser_class(_pattern, self.custom_types)

    def check_match(self, step):
        """
        Check if step matches definition and also for duplicate step definitions.
        Called by behave.

        :type step: str
        :param step: Step definition to be matched

        :return: List of matched arguments if `step` matches, `None` otherwise.
        """

        args = []
        # Escaped quotes and append space (because patterns end with space).
        step_suffix = self.start_phrase + step.replace('\\"', "''") + " "
        # To calculate positions in step argument - used for keyword highlights.
        offset = -len(self.start_phrase)

        def fix_escape_quotes(value, text_repr):
            """
            Replace two consecutive single quotes ('') back to double quote.
            Do not escape double quotes.

            :type value: <any>
            :param value: If instance of str fix quotes.

            :type text_repr: str
            :param text_repr: Text representation of parameter value.

            :return: Tuple of fixed value and text_repr.
            """

            if isinstance(value, str):
                value = value.replace("''", '"')
            text_repr = text_repr.replace("''", '"')
            return value, text_repr

        def process_result(step_suffix, offset, args, result, pattern):
            """
            Convert parse results into behave arguments.

            :type step_suffix: str
            :param step_suffix: Current prefix of step.

            :type offset: int
            :param offset: Current offset (length of already parsed step prefix).

            :type args: list
            :param args: List of already parsed <Argument>s.

            :type result: <parse.Result>
            :param result: Result of current match.

            :type pattern: str
            :param pattern: Currently matched pattern.

            :return: Tuple of updated step_suffix, offset, args.
            """

            for index, value in enumerate(result.fixed):
                start, end = result.spans[index]
                value, text_repr = fix_escape_quotes(value, step_suffix[start:end])
                args.append(Argument(start + offset, end + offset, text_repr, value))
            for name, value in result.named.items():
                start, end = result.spans[name]
                value, text_repr = fix_escape_quotes(value, step_suffix[start:end])
                args.append(
                    Argument(start + offset, end + offset, text_repr, value, name)
                )
            # Matched part of the step.
            pattern_filled = pattern.format_map(result.named)
            # Remove matched part and append start_phrase.
            old_len = len(step_suffix)
            step_suffix = self.start_phrase + step_suffix.replace(
                pattern_filled, ""
            ).lstrip(" |")
            new_len = len(step_suffix)
            # Calculate offset.
            offset += old_len - new_len
            return (step_suffix, offset, args)

        # Check that the first part is first, finish if step does not match.
        pattern = self.patterns[0]
        if len(self.patterns) == 1:
            result = self.parsers[pattern].parse(step_suffix)
        else:
            result = self.parsers[pattern].search(step_suffix)

        if result is None:
            return None
        step_suffix, offset, args = process_result(
            step_suffix, offset, args, result, pattern
        )

        # Match rest of the step.
        patterns = self.patterns[1:]
        while len(step_suffix) > len(self.start_phrase):
            found = False
            for pattern in patterns:
                result = self.parsers[pattern].search(step_suffix)
                if result is None:
                    continue
                # Remove pattern as it was used.
                patterns.remove(pattern)
                step_suffix, offset, args = process_result(
                    step_suffix, offset, args, result, pattern
                )
                found = True
            if not found:
                return None

        args.sort(key=lambda x: x.start)
        return args



#!/usr/bin/env python3
import os
import setuptools


def read_file(filename) -> str:
    """
    Return the content of a file.
    """

    with open(filename, "r", encoding="utf-8") as file:
        return file.read()


def scripts() -> list:
    """
    Return list of scripts in scripts/ file

    :return: List of scripts.
    :rtype: list
    """

    scripts_list = os.listdir(os.curdir + "/scripts/")
    result = []
    for file in scripts_list:
        result = result + ["scripts/" + file]
    return result


setuptools.setup(
    name="qecore",
    version="3.34.2",
    author="Michal Odehnal",
    author_email="modehnal@redhat.com",
    license="GPLv3",
    url="https://gitlab.com/dogtail/qecore",
    description="DesktopQE Tool for unified test execution",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    scripts=scripts(),
    install_requires=[
        "behave<=1.2.6",
        "behave-html-formatter<=0.9.10",
        "behave-html-pretty-formatter",
        "dasbus<=1.7",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Testing :: BDD",
    ],
    python_requires=">=3.6",
    options={
        "build_scripts": {
            "executable": "/usr/bin/python3",
        }
    },
)
